<ul class="navbar-nav">
  <li class="version-switcher__container dropdown">
    <button
      type="button"
      role="button"
      class="version-switcher__button btn btn-sm navbar-btn"
      data-bs-toggle="dropdown"
      id="dLabelMore"
      aria-haspopup="listbox"
    >
      v{{ version }}
      <span class="caret"></span>
    </button>
  </li>
</ul>
